/*
 * init.c
 *
 *        Copyright (C) 2000, 2001  Sony Computer Entertainment Inc.
 *
 * This file is subject to the terms and conditions of the GNU General
 * Public License Version 2. See the file "COPYING" in the main
 * directory of this archive for more details.
 *
 * $Id: init.c,v 1.1 2001/03/28 07:47:23 nakamura Exp $
 */
#include <linux/init.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/types.h>
#include <asm/addrspace.h>
#include <asm/bootinfo.h>
#include <asm/ps2/sbcall.h>

#define SBIOS_ENTRY	0x80001000
#define SBIOS_SIGNATURE	0x80001004

#define CMDLINE		(char **)0x81fff004

char arcs_cmdline[256] = "root=/dev/hda1";

int __init prom_init(int argc, char **argv, char **envp)
{
	char *p, *q;
	int len;

	mips_machgroup = MACH_GROUP_EE;

#ifdef CONFIG_T10000
	mips_machtype = MACH_T10000;
#else /* CONFIG_T10000 */
	mips_machtype = MACH_PS2;
#endif /* !CONFIG_T10000 */

#ifdef CONFIG_T10000_MAXMEM
	mips_memory_upper = 128 * 1024 * 1024 - 4096;	/* 128MB - 4KB */
#else /* CONFIG_T10000_MAXMEM */
	mips_memory_upper = 32 * 1024 * 1024 - 4096;	/* 32MB - 4KB */
#endif /* !CONFIG_T10000_MAXMEM */

	mips_memory_upper |= KSEG0;

	/* get command line parameters */
	if ((p = *CMDLINE) != '\0') {
	    q = arcs_cmdline;
	    len = 0;
	    while ((*q++ = *p++) != '\0') {
		if (len++ > 255) {
		    *q++ = '\0';
		    break;
		}
	    }
	}

	if (*(unsigned long *)SBIOS_SIGNATURE != 0x62325350) {
		/* SBIOS not found */
		while (1)
			;
	}
	sbios = *(int (**)(int, void *))SBIOS_ENTRY;
	return 0;
}

void __init prom_fixup_mem_map(unsigned long start, unsigned long end)
{
	/* nothing to do */
}

void prom_free_prom_memory (void)
{
	/* nothing to do */
}
